.nolist
include \masm32\include\masm32rt.inc

.686
.xmm

.code


;#############################################################################
; Instructions detection code by Alex aka Antariy
OPTION PROLOGUE:NONE
OPTION EPILOGUE:NONE

AxCPUid_Features	struct DWORD
	Is486capable	dd	?
	IsP1		dd	?
	IsP1MMX		dd	?
	IsPPro		dd	?
	IsSSE1		dd	?
	IsSSE2		dd	?
	IsSSE3		dd	?
	IsSSSE3		dd	?
	IsSSE41		dd	?
	IsSSE42		dd	?
	BrandName	db	64 dup (?)
AxCPUid_Features	ends

AxCPUidCodeSizeStart EQU $


; Return values:
; Not zero - if entire structure was filled
; 0 - if CPUID is not supported or not entire struct beed filled.
;     For clearness, check the Is486capable - if 0 then CPUID is not supported,
;     otherwise CPU family less than PIV, but structure are filled properly.

align 16
AxCPUid_FillStructure	proc	lpAxCPUid_Features:PTR AxCPUid_Features
	.data?
		ifrunnedAndSupportCPUid	dd	?
	.code
	
	push ebp
	push ebx
	
	xor eax,eax
	mov ecx,sizeof AxCPUid_Features

	mov ebp,[esp+4+8]
	
	@@:
	mov [ebp+ecx-4].AxCPUid_Features.Is486capable,eax
	add ecx,-4
	jnz @B

	mov ebx,ifrunnedAndSupportCPUid
	dec ebx
	jz @done
	dec ebx
	jz @is486capable
	
	inc ifrunnedAndSupportCPUid
	
	pushfd
	pop ecx
	mov ebx,ecx
	xor ebx,200000h	; switch ID flag
	push ebx
	popfd		; change EFLAGS
	pushfd
	pop ebx
	xor ebx,ecx	; if previous ID bit was equal to current
	test ebx,200000h; then it would be dropped
	jz @done	; otherwise - it set
	
	inc ifrunnedAndSupportCPUid
	
	@is486capable:

	or [ebp].AxCPUid_Features.Is486capable,1
	
	xor eax,eax
	cpuid
	test eax,eax
	jz @done
	
	mov eax,1
	cpuid
	
	shr edx,5	; has RDTSC - so, this is PI at least
	adc [ebp].AxCPUid_Features.IsP1,0
	
	shr edx,2	; has PAE - so, PPro at least
	adc [ebp].AxCPUid_Features.IsPPro,0
	
	shr edx,17	; has MMX
	adc [ebp].AxCPUid_Features.IsP1MMX,0
	
	shr edx,2	; has SSE1
	adc [ebp].AxCPUid_Features.IsSSE1,0
	
	shr edx,1	; has SSE2
	adc [ebp].AxCPUid_Features.IsSSE2,0
	
	
	
	shr ecx,1	; has SSE3
	adc [ebp].AxCPUid_Features.IsSSE3,0
	
	shr ecx,9	; has SSSE3
	adc [ebp].AxCPUid_Features.IsSSSE3,0
	
	shr ecx,10	; has SSE4.1
	adc [ebp].AxCPUid_Features.IsSSE41,0
	
	shr ecx,1	; has SSE4.2
	adc [ebp].AxCPUid_Features.IsSSE42,0
	
	
	; get CPU brand name, if exist
	
	mov eax,80000000h
	cpuid
	pushad
	print hex$(eax),9,"Debug message: extended functions count return",13,10,13,10
	popad
	add eax,eax	; check for zero and no any extended functions
	jz @done
	
	; temporary fix for PIII, not SSE2 capable CPU cannot have brand name
	cmp [ebp].AxCPUid_Features.IsSSE2,0
	jz @done
	
	lea ebp,[ebp].AxCPUid_Features.BrandName
	cmp eax,8	; needed at least 80000004h function
	mov eax,0
	jb @done
	
	push esi
	mov esi,80000002h
	@@:
	mov eax,esi
	cpuid
	inc esi
	mov [ebp],eax
	mov [ebp+4],ebx
	mov [ebp+8],ecx
	mov [ebp+12],edx
	add ebp,16
	cmp esi,80000005h
	jb @B		
	pop esi
	
	or eax,1	; in case of terminated brand string
	
	@done:
	pop ebx
	pop ebp
	ret 4
AxCPUid_FillStructure	endp


; Return values:
; 0 - need upgrade
; above 0 - then supported:
; 1 - MMX
; 2 - SSE1
; 3 - SSE2
; 4 - SSE3
; 5 - SSSE3
; 6 - SSE4.1
; 7 - SSE4.2
align 16
AxCPUid_Print	proc
	push ebx
	push esi
	push edi
	add esp,-1028
	
	invoke GetStdHandle,STD_OUTPUT_HANDLE
	xchg eax,ebx
		
	invoke AxCPUid_FillStructure,esp
	cmp [esp].AxCPUid_Features.Is486capable,0
	jnz @F	
	
	push eax
	mov edx,esp
	
	push 0
	push edx
	push sizeof @@needupgrade
	push offset @@needupgrade
	push ebx
	call WriteFile
	pop ecx
	xor eax,eax
	jmp @done
	
	@@:
	
	mov esi,esp
	
	mov edi,[esi].AxCPUid_Features.IsP1MMX
	add edi,[esi].AxCPUid_Features.IsSSE1
	add edi,[esi].AxCPUid_Features.IsSSE2
	add edi,[esi].AxCPUid_Features.IsSSE3
	add edi,[esi].AxCPUid_Features.IsSSSE3
	add edi,[esi].AxCPUid_Features.IsSSE41
	add edi,[esi].AxCPUid_Features.IsSSE42
		
	mov eax,[esi].AxCPUid_Features.IsSSE42
	lea eax,[eax*2+offset @@sse42]

	mov edx,[esi].AxCPUid_Features.IsSSE41
	lea edx,[edx*2+offset @@sse41]
	
	mov ecx,[esi].AxCPUid_Features.IsSSSE3
	lea ecx,[ecx*2+offset @@ssse3]
	
	push eax
	push edx
	push ecx
	
	mov eax,[esi].AxCPUid_Features.IsSSE3
	lea eax,[eax*2+offset @@sse3]
	
	mov edx,[esi].AxCPUid_Features.IsSSE2
	lea edx,[edx*2+offset @@sse2]
	
	mov ecx,[esi].AxCPUid_Features.IsSSE1
	lea ecx,[ecx*2+offset @@sse1]
	
	push eax
	push edx
	push ecx	
	
	mov eax,[esi].AxCPUid_Features.IsP1MMX
	lea eax,[eax*2+offset @@mmx]
	
	lea edx,[esi].AxCPUid_Features.BrandName
	cmp dword ptr [edx],0
	jnz @hasbrandname

	mov ecx,[esi].AxCPUid_Features.Is486capable
	add ecx,[esi].AxCPUid_Features.IsP1
	add ecx,[esi].AxCPUid_Features.IsP1MMX
	add ecx,[esi].AxCPUid_Features.IsPPro
	
	cmp ecx,4
	jb @F
	add ecx,[esi].AxCPUid_Features.IsP1MMX	; PII is PPro with MMX
	add ecx,[esi].AxCPUid_Features.IsSSE1	; PIII is PPro with MMX and SSE1
	@@:
		
	mov edx,[ecx*4+offset @@earlycpus]
		
	@hasbrandname:
	add edx,1
	cmp byte ptr [edx-1]," "
	jz @hasbrandname
	
	dec edx
		
	push eax
	push edx
	
	push offset @@fmtFeatures
	push esi
	call wsprintf

	mov edx,esp
	
	invoke WriteFile,ebx,esi,eax,edx,0
	
	add esp,10*4

	xchg eax,edi	

	@done:
	add esp,1028
	pop edi
	pop esi
	pop ebx
	ret
	
	even
	@@needupgrade	db	"This is time for upgrade indeed, i386 or early i486 :)"
	even
	@@i486		db	"Old-good 80486",0
	even
	@@p1		db	"Pentium",0
	even
	@@pmmx		db	"Pentium with MMX Technology",0
	even
	@@ppro		db	"Pentium Pro",0
	even
	@@p2		db	"Pentium II",0
	even
	@@p3		db	"Pentium III",0
	
	align 4
	@@earlycpus	dd	0
			dd	offset @@i486
			dd	offset @@p1
			dd	offset @@pmmx
			dd	offset @@ppro
			dd	offset @@p2
			dd	offset @@p3
	
	even
	@@fmtFeatures	db	"%s",13,10,13,10
			db	"Instructions: %s%s%s%s%s%s%s",13,10,13,10,0
	even
	@@mmx		db	0,0
			db	"MMX",0
	even
	@@sse1		db	0,0
			db	", SSE1",0
	even
	@@sse2		db	0,0
			db	", SSE2",0
	even
	@@sse3		db	0,0
			db	", SSE3",0
	even
	@@ssse3		db	0,0
			db	", SSSE3",0
	even
	@@sse41		db	0,0
			db	", SSE4.1",0
	even
	@@sse42		db	0,0
			db	", SSE4.2",0
	
	
AxCPUid_Print	endp
AxCPUidCodeSize EQU $-AxCPUidCodeSizeStart

OPTION PROLOGUE:PrologueDef
OPTION EPILOGUE:EpilogueDef
;#############################################################################

.code
start:

invoke AxCPUid_Print
print chr$(13,10,13,10)
mov eax,AxCPUidCodeSize
print str$(eax)," - size of all code and data of AxCPUid filling and printing proc",13,10,13,10
inkey
exit

end start
